from autolens.pipeline import phase_tagging


class TestPhaseTag:
    def test__mixture_of_values(self):

        phase_tag = phase_tagging.phase_tag_from_phase_settings(
            sub_grid_size=2,
            signal_to_noise_limit=2,
            bin_up_factor=None,
            image_psf_shape=None,
            inversion_psf_shape=None,
            inner_mask_radii=0.3,
            positions_threshold=2.0,
            interp_pixel_scale=None,
            cluster_pixel_scale=None,
        )

        assert phase_tag == "_sub_2_snr_2_pos_2.00_inner_mask_0.30"

        phase_tag = phase_tagging.phase_tag_from_phase_settings(
            sub_grid_size=1,
            signal_to_noise_limit=None,
            bin_up_factor=3,
            image_psf_shape=(2, 2),
            inversion_psf_shape=(3, 3),
            inner_mask_radii=None,
            positions_threshold=None,
            interp_pixel_scale=0.2,
            cluster_pixel_scale=0.3,
        )

        assert (
            phase_tag
            == "_sub_1_bin_3_image_psf_2x2_inv_psf_3x3_interp_0.200_cluster_0.300"
        )


class TestPhaseTaggers:
    def test__positions_threshold_tagger(self):

        tag = phase_tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=None
        )
        assert tag == ""
        tag = phase_tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=1.0
        )
        assert tag == "_pos_1.00"
        tag = phase_tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=2.56
        )
        assert tag == "_pos_2.56"

    def test__inner_circular_mask_radii_tagger(self):

        tag = phase_tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=None
        )
        assert tag == ""
        tag = phase_tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=0.2
        )
        print(tag)
        assert tag == "_inner_mask_0.20"
        tag = phase_tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=3
        )
        assert tag == "_inner_mask_3.00"

    def test__sub_grid_size_tagger(self):

        tag = phase_tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=1)
        assert tag == "_sub_1"
        tag = phase_tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=2)
        assert tag == "_sub_2"
        tag = phase_tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=4)
        assert tag == "_sub_4"

    def test__signal_to_noise_limit_tagger(self):

        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=None
        )
        assert tag == ""
        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=1
        )
        assert tag == "_snr_1"
        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=2
        )
        assert tag == "_snr_2"
        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=3
        )
        assert tag == "_snr_3"

    def test__bin_up_factor_tagger(self):

        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=None)
        assert tag == ""
        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=1)
        assert tag == ""
        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=2)
        assert tag == "_bin_2"
        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=3)
        assert tag == "_bin_3"

    def test__image_psf_shape_tagger(self):

        tag = phase_tagging.image_psf_shape_tag_from_image_psf_shape(
            image_psf_shape=None
        )
        assert tag == ""
        tag = phase_tagging.image_psf_shape_tag_from_image_psf_shape(
            image_psf_shape=(2, 2)
        )
        assert tag == "_image_psf_2x2"
        tag = phase_tagging.image_psf_shape_tag_from_image_psf_shape(
            image_psf_shape=(3, 4)
        )
        assert tag == "_image_psf_3x4"

    def test__inversion_psf_shape_tagger(self):

        tag = phase_tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=None
        )
        assert tag == ""
        tag = phase_tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=(2, 2)
        )
        assert tag == "_inv_psf_2x2"
        tag = phase_tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=(3, 4)
        )
        assert tag == "_inv_psf_3x4"

    def test__interp_pixel_scale_tagger(self):

        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=None
        )
        assert tag == ""
        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.5
        )
        assert tag == "_interp_0.500"
        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.25
        )
        assert tag == "_interp_0.250"
        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.234
        )
        assert tag == "_interp_0.234"

    def test__cluster_pixel_scale_tagger(self):

        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=None
        )
        assert tag == ""
        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.5
        )
        assert tag == "_cluster_0.500"
        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.25
        )
        assert tag == "_cluster_0.250"
        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.234
        )
        assert tag == "_cluster_0.234"



import os
import shutil

import numpy as np
import pytest

from autolens import exc
from autolens.data.array import scaled_array
from autolens.data.instrument import abstract_data
from autolens.data.instrument import interferometer
from autolens.data.array.util import grid_util, mapping_util

test_data_dir = "{}/../../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestInterferometerData(object):
    
    class TestNewInterferometerDataResized:

        def test__all_components_resized__psf_and_primary_beam_are_not(self):
            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            image_array[3, 3] = 2.0

            noise_map_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            noise_map_array[3, 3] = 3.0

            exposure_time_map_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            exposure_time_map_array[3, 3] = 5.0

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                primary_beam=interferometer.PrimaryBeam(np.zeros((5,5)), pixel_scale=1.0),
                noise_map=noise_map_array,
                exposure_time_map=exposure_time_map_array,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_with_resized_arrays(new_shape=(4, 4))

            assert (
                    interferometer_data.image
                    == np.array(
                [
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 2.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ]
            )
            ).all()
            assert (
                    interferometer_data.noise_map
                    == np.array(
                [
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 3.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ]
            )
            ).all()

            assert (
                    interferometer_data.exposure_time_map
                    == np.array(
                [
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 5.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0],
                ]
            )
            ).all()

            assert interferometer_data.pixel_scale == 1.0
            assert (interferometer_data.psf == np.zeros((3, 3))).all()
            assert (interferometer_data.primary_beam == np.zeros((5, 5))).all()
            assert interferometer_data.origin == (0.0, 0.0)

        def test__resize_psf(self):
            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                primary_beam=interferometer.PrimaryBeam(np.zeros((5,5)), pixel_scale=1.0),
                noise_map=1,
                exposure_time_map=1,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_with_resized_psf(new_shape=(1, 1))

            assert (interferometer_data.image == np.ones((6, 6))).all()
            assert interferometer_data.pixel_scale == 1.0
            assert (interferometer_data.psf == np.zeros((1, 1))).all()
            assert interferometer_data.origin == (0.0, 0.0)
            
        def test__resize_primary_beam(self):
            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=1,
                primary_beam=interferometer.PrimaryBeam(np.zeros((5,5)), pixel_scale=1.0),
                noise_map=1,
                exposure_time_map=1,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_with_resized_primary_beam(new_shape=(1, 1))

            assert (interferometer_data.image == np.ones((6, 6))).all()
            assert interferometer_data.pixel_scale == 1.0
            assert (interferometer_data.primary_beam == np.zeros((1, 1))).all()
            assert interferometer_data.origin == (0.0, 0.0)

        def test__input_new_centre_pixels__arrays_use_new_centre__psf_and_primary_beam_do_not(self):
            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            image_array[3, 3] = 2.0

            noise_map_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            noise_map_array[3, 3] = 3.0

            exposure_time_map_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            exposure_time_map_array[3, 3] = 5.0

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                primary_beam=interferometer.PrimaryBeam(np.zeros((5, 5)), pixel_scale=1.0),
                noise_map=noise_map_array,
                exposure_time_map=exposure_time_map_array,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_with_resized_arrays(
                new_shape=(3, 3), new_centre_pixels=(3, 3)
            )

            assert (
                    interferometer_data.image
                    == np.array([[1.0, 1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 1.0, 1.0]])
            ).all()
            assert (
                    interferometer_data.noise_map
                    == np.array([[1.0, 1.0, 1.0], [1.0, 3.0, 1.0], [1.0, 1.0, 1.0]])).all()

            assert (
                    interferometer_data.exposure_time_map
                    == np.array([[1.0, 1.0, 1.0], [1.0, 5.0, 1.0], [1.0, 1.0, 1.0]])
            ).all()

            assert interferometer_data.pixel_scale == 1.0
            assert (interferometer_data.psf == np.zeros((3, 3))).all()
            assert (interferometer_data.primary_beam == np.zeros((5, 5))).all()
            assert interferometer_data.origin == (0.0, 0.0)

        def test__input_new_centre_arcsec__arrays_use_new_centre__psf_and_primary_beam_do_not(self):
            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            image_array[3, 3] = 2.0

            noise_map_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            noise_map_array[3, 3] = 3.0

            exposure_time_map_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            exposure_time_map_array[3, 3] = 5.0

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                primary_beam=interferometer.PrimaryBeam(np.zeros((5,5)), pixel_scale=1.0),
                noise_map=noise_map_array,
                exposure_time_map=exposure_time_map_array,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_with_resized_arrays(
                new_shape=(3, 3), new_centre_arcsec=(-0.5, 0.5)
            )

            assert (
                    interferometer_data.image
                    == np.array([[1.0, 1.0, 1.0], [1.0, 2.0, 1.0], [1.0, 1.0, 1.0]])
            ).all()
            assert (
                    interferometer_data.noise_map
                    == np.array([[1.0, 1.0, 1.0], [1.0, 3.0, 1.0], [1.0, 1.0, 1.0]])
            ).all()
            assert (
                    interferometer_data.exposure_time_map
                    == np.array([[1.0, 1.0, 1.0], [1.0, 5.0, 1.0], [1.0, 1.0, 1.0]])
            ).all()

            assert interferometer_data.pixel_scale == 1.0
            assert (interferometer_data.psf == np.zeros((3, 3))).all()
            assert (interferometer_data.primary_beam == np.zeros((5, 5))).all()
            assert interferometer_data.origin == (0.0, 0.0)

        def test__input_both_centres__raises_error(self):
            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((6, 6)), pixel_scale=1.0
            )
            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                primary_beam=interferometer.PrimaryBeam(np.zeros((5, 5)), pixel_scale=1.0),
                noise_map=1,
                exposure_time_map=1,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            with pytest.raises(exc.DataException):
                interferometer_data.new_interferometer_data_with_resized_arrays(
                    new_shape=(3, 3),
                    new_centre_pixels=(3, 3),
                    new_centre_arcsec=(-0.5, 0.5),
                )

    class TestNewImageConvertedFrom:
        def test__counts__all_arrays_in_units_of_flux_are_converted(self):

            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((3, 3)), pixel_scale=1.0
            )
            noise_map_array = scaled_array.ScaledSquarePixelArray(
                2.0 * np.ones((3, 3)), pixel_scale=1.0
            )
            exposure_time_map_array = scaled_array.ScaledSquarePixelArray(
                0.5 * np.ones((3, 3)), pixel_scale=1.0
            )

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                noise_map=noise_map_array,
                exposure_time_map=exposure_time_map_array,
                primary_beam=1,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_converted_from_electrons()

            assert (interferometer_data.image == 2.0 * np.ones((3, 3))).all()
            assert (interferometer_data.noise_map == 4.0 * np.ones((3, 3))).all()
            assert interferometer_data.origin == (0.0, 0.0)

        def test__adus__all_arrays_in_units_of_flux_are_converted(self):

            image_array = scaled_array.ScaledSquarePixelArray(
                np.ones((3, 3)), pixel_scale=1.0
            )
            noise_map_array = scaled_array.ScaledSquarePixelArray(
                2.0 * np.ones((3, 3)), pixel_scale=1.0
            )
            exposure_time_map_array = scaled_array.ScaledSquarePixelArray(
                0.5 * np.ones((3, 3)), pixel_scale=1.0
            )

            interferometer_data = interferometer.InteferometerData(
                image=image_array,
                pixel_scale=1.0,
                psf=abstract_data.PSF(np.zeros((3, 3)), pixel_scale=1.0),
                noise_map=noise_map_array,
                exposure_time_map=exposure_time_map_array,
                primary_beam=1,
                real_visibilities=1, imaginary_visibilities=1, visibilities_noise_map=1, u_wavelengths=1, v_wavelengths=1,
            )

            interferometer_data = interferometer_data.new_interferometer_data_converted_from_adus(gain=2.0)

            assert (interferometer_data.image == 2.0 * 2.0 * np.ones((3, 3))).all()
            assert (interferometer_data.noise_map == 2.0 * 4.0 * np.ones((3, 3))).all()
            assert interferometer_data.origin == (0.0, 0.0)


class TestPrimaryBeam(object):
    
    class TestConstructors(object):

        def test__init__input_primary_beam__all_attributes_correct_including_data_inheritance(
            self
        ):
            psf = interferometer.PrimaryBeam(array=np.ones((3, 3)), pixel_scale=1.0, renormalize=False)

            assert psf.shape == (3, 3)
            assert psf.pixel_scale == 1.0
            assert (psf == np.ones((3, 3))).all()
            assert psf.origin == (0.0, 0.0)

            psf = interferometer.PrimaryBeam(array=np.ones((4, 3)), pixel_scale=1.0, renormalize=False)

            assert (psf == np.ones((4, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.shape == (4, 3)
            assert psf.origin == (0.0, 0.0)

        def test__from_fits__input_primary_beam_3x3__all_attributes_correct_including_data_inheritance(
            self
        ):
            psf = interferometer.PrimaryBeam.from_fits_with_scale(
                file_path=test_data_dir + "3x3_ones.fits", hdu=0, pixel_scale=1.0
            )

            assert (psf == np.ones((3, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.origin == (0.0, 0.0)

            psf = interferometer.PrimaryBeam.from_fits_with_scale(
                file_path=test_data_dir + "4x3_ones.fits", hdu=0, pixel_scale=1.0
            )

            assert (psf == np.ones((4, 3))).all()
            assert psf.pixel_scale == 1.0
            assert psf.origin == (0.0, 0.0)

    class TestRenormalize(object):
        def test__input_is_already_normalized__no_change(self):
            primary_beam_data = np.ones((3, 3)) / 9.0

            psf = interferometer.PrimaryBeam(array=primary_beam_data, pixel_scale=1.0, renormalize=True)

            assert psf == pytest.approx(primary_beam_data, 1e-3)

        def test__input_is_above_normalization_so_is_normalized(self):

            primary_beam_data = np.ones((3, 3))

            psf = interferometer.PrimaryBeam(array=primary_beam_data, pixel_scale=1.0, renormalize=True)

            assert psf == pytest.approx(np.ones((3, 3)) / 9.0, 1e-3)

        def test__same_as_above__renomalized_false_does_not_renormalize(self):
            primary_beam_data = np.ones((3, 3))

            psf = interferometer.PrimaryBeam(array=primary_beam_data, pixel_scale=1.0, renormalize=False)

            assert psf == pytest.approx(np.ones((3, 3)), 1e-3)

    class TestBinnedUp(object):
        def test__primary_beam_is_even_x_even__rescaled_to_odd_x_odd__no_use_of_dimension_trimming(
            self
        ):

            array_2d = np.ones((6, 6))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((9, 9))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.333333333333333, renormalize=True
            )
            assert psf.pixel_scale == 3.0
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((18, 6))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 27.0) * np.ones((9, 3))

            array_2d = np.ones((6, 18))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == 2.0
            assert psf == (1.0 / 27.0) * np.ones((3, 9))

        def test__primary_beam_is_even_x_even_after_binning_up__resized_to_odd_x_odd_with_shape_plus_one(
            self
        ):

            array_2d = np.ones((2, 2))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == 0.4
            assert psf == (1.0 / 25.0) * np.ones((5, 5))

            array_2d = np.ones((40, 40))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.1, renormalize=True
            )
            assert psf.pixel_scale == 8.0
            assert psf == (1.0 / 25.0) * np.ones((5, 5))

            array_2d = np.ones((2, 4))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(0.4444444, 1.0e-4)
            assert psf == (1.0 / 45.0) * np.ones((5, 9))

            array_2d = np.ones((4, 2))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=2.0, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(0.4444444, 1.0e-4)
            assert psf == (1.0 / 45.0) * np.ones((9, 5))

        def test__primary_beam_is_odd_and_even_after_binning_up__resized_to_odd_and_odd_with_shape_plus_one(
            self
        ):

            array_2d = np.ones((6, 4))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(2.0, 1.0e-4)
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((9, 12))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.33333333333, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(3.0, 1.0e-4)
            assert psf == (1.0 / 15.0) * np.ones((3, 5))

            array_2d = np.ones((4, 6))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.5, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(2.0, 1.0e-4)
            assert psf == (1.0 / 9.0) * np.ones((3, 3))

            array_2d = np.ones((12, 9))
            psf = interferometer.PrimaryBeam(array=array_2d, pixel_scale=1.0, renormalize=False)
            psf = psf.new_primary_beam_with_rescaled_odd_dimensioned_array(
                rescale_factor=0.33333333333, renormalize=True
            )
            assert psf.pixel_scale == pytest.approx(3.0, 1.0e-4)
            assert psf == (1.0 / 15.0) * np.ones((5, 3))

    class TestNewRenormalizedPrimaryBeam(object):
        def test__input_is_already_normalized__no_change(self):

            primary_beam_data = np.ones((3, 3)) / 9.0

            psf = interferometer.PrimaryBeam(array=primary_beam_data, pixel_scale=1.0, renormalize=False)

            primary_beam_new = psf.new_primary_beam_with_renormalized_array()

            assert primary_beam_new == pytest.approx(primary_beam_data, 1e-3)

        def test__input_is_above_normalization_so_is_normalized(self):

            primary_beam_data = np.ones((3, 3))

            psf = interferometer.PrimaryBeam(array=primary_beam_data, pixel_scale=1.0, renormalize=False)

            primary_beam_new = psf.new_primary_beam_with_renormalized_array()

            assert primary_beam_new == pytest.approx(np.ones((3, 3)) / 9.0, 1e-3)

    class TestConvolve(object):
        def test__kernel_is_not_odd_x_odd__raises_error(self):
            kernel = np.array([[0.0, 1.0], [1.0, 2.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            with pytest.raises(exc.KernelException):
                psf.convolve(np.ones((5, 5)))

        def test__image_is_3x3_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (blurred_image == kernel).all()

        def test__image_is_4x4_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [0.0, 1.0, 0.0, 0.0],
                        [1.0, 2.0, 1.0, 0.0],
                        [0.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0],
                    ]
                )
            ).all()

        def test__image_is_4x3_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]]
                )
            ).all()

        def test__image_is_3x4_central_value_of_one__kernel_is_cross__blurred_image_becomes_cross(
            self
        ):
            image = np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )

            kernel = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [[0.0, 1.0, 0.0, 0.0], [1.0, 2.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0]]
                )
            ).all()

        def test__image_is_4x4_has_two_central_values__kernel_is_asymmetric__blurred_image_follows_convolution(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [1.0, 1.0, 1.0, 0.0],
                        [2.0, 3.0, 2.0, 1.0],
                        [1.0, 5.0, 5.0, 1.0],
                        [0.0, 1.0, 3.0, 3.0],
                    ]
                )
            ).all()

        def test__image_is_4x4_values_are_on_edge__kernel_is_asymmetric__blurring_does_not_account_for_edge_effects(
            self
        ):
            image = np.array(
                [
                    [0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [1.0, 1.0, 0.0, 0.0],
                        [2.0, 1.0, 1.0, 1.0],
                        [3.0, 3.0, 2.0, 2.0],
                        [0.0, 0.0, 1.0, 3.0],
                    ]
                )
            ).all()

        def test__image_is_4x4_values_are_on_corner__kernel_is_asymmetric__blurring_does_not_account_for_edge_effects(
            self
        ):
            image = np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )

            kernel = np.array([[1.0, 1.0, 1.0], [2.0, 2.0, 1.0], [1.0, 3.0, 3.0]])

            psf = interferometer.PrimaryBeam(array=kernel, pixel_scale=1.0)

            blurred_image = psf.convolve(image)

            assert (
                blurred_image
                == np.array(
                    [
                        [2.0, 1.0, 0.0, 0.0],
                        [3.0, 3.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 1.0],
                        [0.0, 0.0, 2.0, 2.0],
                    ]
                )
            ).all()

    class TestFromGaussian(object):
        def test__identical_to_gaussian_light_profile(self):

            from autolens.model.profiles import light_profiles as lp

            grid = grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
                mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_grid_size=1
            )

            gaussian = lp.EllipticalGaussian(
                centre=(0.1, 0.1), axis_ratio=0.9, phi=45.0, intensity=1.0, sigma=1.0
            )
            profile_gaussian_1d = gaussian.intensities_from_grid(grid)
            profile_gaussian_2d = mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
                sub_array_1d=profile_gaussian_1d,
                mask=np.full(fill_value=False, shape=(3, 3)),
                sub_grid_size=1,
            )
            profile_psf = interferometer.PrimaryBeam(
                array=profile_gaussian_2d, pixel_scale=1.0, renormalize=True
            )

            imaging_psf = interferometer.PrimaryBeam.from_gaussian(
                shape=(3, 3),
                pixel_scale=1.0,
                centre=(0.1, 0.1),
                axis_ratio=0.9,
                phi=45.0,
                sigma=1.0,
            )

            assert profile_psf == pytest.approx(imaging_psf, 1e-4)
            
            
class TestInterferometerFromFits(object):
    
    def test__no_settings_just_pass_fits(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            visibilities_noise_map_path=test_data_dir + '3_threes.fits',
            u_wavelengths_path=test_data_dir + '3_fours.fits',
            v_wavelengths_path=test_data_dir + '3_fives.fits',
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (interferometer_data.real_visibilities == np.ones(3)).all()
        assert (interferometer_data.imaginary_visibilities == 2.0 * np.ones(3)).all()
        assert (interferometer_data.visibilities_noise_map == 3.0 * np.ones(3)).all()
        assert (interferometer_data.u_wavelengths == 4.0 * np.ones(3)).all()
        assert (interferometer_data.v_wavelengths == 5.0 * np.ones(3)).all()
        assert (interferometer_data.primary_beam == 5.0 * np.ones((3,3))).all()

        assert interferometer_data.pixel_scale == 0.1

    def test__optional_array_paths_included__loads_optional_array(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (interferometer_data.primary_beam == 5.0 * np.ones((3,3))).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1

    def test__all_files_in_one_fits__load_using_different_hdus(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_multiple_hdu.fits",
            image_hdu=0,
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_multiple_hdu.fits",
            psf_hdu=1,
            noise_map_path=test_data_dir + "3x3_multiple_hdu.fits",
            noise_map_hdu=2,
            primary_beam_path=test_data_dir + "3x3_multiple_hdu.fits",
            primary_beam_hdu=3,
            exposure_time_map_path=test_data_dir + "3x3_multiple_hdu.fits",
            exposure_time_map_hdu=5,
            real_visibilities_path=test_data_dir + '3_multiple_hdu.fits',
            real_visibilities_hdu=0,
            imaginary_visibilities_path=test_data_dir + '3_multiple_hdu.fits',
            imaginary_visibilities_hdu=1,
            visibilities_noise_map_path=test_data_dir + '3_multiple_hdu.fits',
            visibilities_noise_map_hdu=2,
            u_wavelengths_path=test_data_dir + '3_multiple_hdu.fits',
            u_wavelengths_hdu=3,
            v_wavelengths_path=test_data_dir + '3_multiple_hdu.fits',
            v_wavelengths_hdu=4,
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (interferometer_data.primary_beam == 4.0 * np.ones((3, 3))).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()
        assert (interferometer_data.real_visibilities == np.ones(3)).all()
        assert (interferometer_data.imaginary_visibilities == 2.0 * np.ones(3)).all()
        assert (interferometer_data.visibilities_noise_map == 3.0 * np.ones(3)).all()
        assert (interferometer_data.u_wavelengths == 4.0 * np.ones(3)).all()
        assert (interferometer_data.v_wavelengths == 5.0 * np.ones(3)).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__exposure_time_included__creates_exposure_time_map_using_exposure_time(
        self
    ):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            noise_map_path=test_data_dir + "3x3_ones.fits",
            psf_path=test_data_dir + "3x3_ones.fits",
            primary_beam_path=test_data_dir + "3x3_ones.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            pixel_scale=0.1,
            exposure_time_map_from_single_value=3.0,
            renormalize_psf=False,
        )

        assert (interferometer_data.exposure_time_map == 3.0 * np.ones((3, 3))).all()

    def test__pad_shape_of_image_arrays_and_psf(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            resized_interferometer_shape=(5, 5),
            resized_psf_shape=(7, 7),
            resized_primary_beam_shape=(9, 9),
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        padded_array = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )


        exposure_padded_array = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 6.0, 6.0, 6.0, 0.0],
                [0.0, 6.0, 6.0, 6.0, 0.0],
                [0.0, 6.0, 6.0, 6.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        psf_padded_array = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        primary_beam_padded_array = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        assert (interferometer_data.image == padded_array).all()
        assert (interferometer_data.psf == psf_padded_array).all()
        assert (interferometer_data.noise_map == 3.0 * padded_array).all()
        assert (interferometer_data.exposure_time_map == 6.0 * padded_array).all()
        assert (interferometer_data.primary_beam == primary_beam_padded_array).all()
        assert (interferometer_data.exposure_time_map == exposure_padded_array).all()

        assert interferometer_data.pixel_scale == 0.1

    def test__trim_shape_of_image_arrays_and_psf(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            resized_interferometer_shape=(1, 1),
            resized_psf_shape=(1, 1),
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        trimmed_array = np.array([[1.0]])

        assert (interferometer_data.image == trimmed_array).all()
        assert (interferometer_data.psf == 2.0 * trimmed_array).all()
        assert (interferometer_data.noise_map == 3.0 * trimmed_array).all()
        assert (interferometer_data.primary_beam == 5.0 * trimmed_array).all()
        assert (interferometer_data.exposure_time_map == 6.0 * trimmed_array).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__convert_noise_map_from_weight_map(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            convert_noise_map_from_weight_map=True,
            renormalize_psf=False,
        )

        noise_map_converted = interferometer.NoiseMap.from_weight_map(
            weight_map=3.0 * np.ones((3, 3)), pixel_scale=0.1
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == noise_map_converted).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__convert_noise_map_from_inverse_noise_map(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            convert_noise_map_from_inverse_noise_map=True,
            renormalize_psf=False,
        )

        noise_map_converted = interferometer.NoiseMap.from_inverse_noise_map(
            inverse_noise_map=3.0 * np.ones((3, 3)), pixel_scale=0.1
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == noise_map_converted).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__psf_renormalized_true__renormalized_psf(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            renormalize_psf=True,
            renormalize_primary_beam=False,
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert interferometer_data.psf == pytest.approx((1.0 / 9.0) * np.ones((3, 3)), 1e-2)
        assert (interferometer_data.primary_beam == 5.0 * np.ones((3,3))).all()
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__primary_beam_renormalized_true__renormalized_psf(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            renormalize_psf=False,
            renormalize_primary_beam=True,
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert interferometer_data.primary_beam == pytest.approx((1.0 / 9.0) * np.ones((3, 3)), 1e-2)
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__convert_image_from_electrons_using_exposure_time(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            renormalize_psf=False,
            renormalize_primary_beam=False,
            convert_from_electrons=True,
        )

        assert (interferometer_data.image == np.ones((3, 3)) / 6.0).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3)) / 6.0).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__convert_image_from_adus_using_exposure_time_and_gain(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            renormalize_psf=False,
            gain=2.0,
            convert_from_adus=True,
        )

        assert (interferometer_data.image == 2.0 * np.ones((3, 3)) / 6.0).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == 2.0 * 3.0 * np.ones((3, 3)) / 6.0).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1

    def test__no_noise_map_input__raises_imaging_exception(self):

        with pytest.raises(exc.DataException):
            interferometer.load_interferometer_data_from_fits(
                image_path=test_data_dir + "3x3_ones.fits",
                pixel_scale=0.1,
                psf_path=test_data_dir + "3x3_twos.fits",
            )

    def test__multiple_noise_map_options__raises_imaging_exception(self):

        with pytest.raises(exc.DataException):
            interferometer.load_interferometer_data_from_fits(
                image_path=test_data_dir + "3x3_ones.fits",
                pixel_scale=0.1,
                psf_path=test_data_dir + "3x3_twos.fits",
                noise_map_path=test_data_dir + "3x3_threes.fits",
                convert_noise_map_from_inverse_noise_map=True,
                convert_noise_map_from_weight_map=True,
            )

    def test__exposure_time_and_exposure_time_map_included__raies_imaging_error(self):

        with pytest.raises(exc.DataException):
            interferometer.load_interferometer_data_from_fits(
                image_path=test_data_dir + "3x3_ones.fits",
                psf_path=test_data_dir + "3x3_ones.fits",
                pixel_scale=0.1,
                noise_map_path=test_data_dir + "3x3_threes.fits",
                exposure_time_map_path=test_data_dir + "3x3_ones.fits",
                exposure_time_map_from_single_value=1.0,
            )

    def test__output_all_arrays(self):

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=test_data_dir + "3x3_ones.fits",
            pixel_scale=0.1,
            psf_path=test_data_dir + "3x3_twos.fits",
            noise_map_path=test_data_dir + "3x3_threes.fits",
            primary_beam_path=test_data_dir + "3x3_fives.fits",
            exposure_time_map_path=test_data_dir + "3x3_sixes.fits",
            real_visibilities_path=test_data_dir + '3_ones.fits',
            imaginary_visibilities_path=test_data_dir + '3_twos.fits',
            visibilities_noise_map_path=test_data_dir + '3_threes.fits',
            u_wavelengths_path=test_data_dir + '3_fours.fits',
            v_wavelengths_path=test_data_dir + '3_fives.fits',
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        output_data_dir = "{}/../test_files/array/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        interferometer.output_interferometer_data_to_fits(
            interferometer_data=interferometer_data,
            image_path=output_data_dir + "image.fits",
            psf_path=output_data_dir + "psf.fits",
            noise_map_path=output_data_dir + "noise_map.fits",
            primary_beam_path=output_data_dir + "primary_beam.fits",
            exposure_time_map_path=output_data_dir + "exposure_time_map.fits",
            real_visibilities_path=output_data_dir + 'real_visibilities.fits',
            imaginary_visibilities_path=output_data_dir + 'imaginary_visibilities.fits',
            visibilities_noise_map_path=output_data_dir + 'visibilities_noise_map.fits',
            u_wavelengths_path=output_data_dir + 'u_wavelengths.fits',
            v_wavelengths_path=output_data_dir + 'v_wavelengths.fits',
            overwrite=True,
        )

        interferometer_data = interferometer.load_interferometer_data_from_fits(
            image_path=output_data_dir + "image.fits",
            pixel_scale=0.1,
            psf_path=output_data_dir + "psf.fits",
            noise_map_path=output_data_dir + "noise_map.fits",
            primary_beam_path=output_data_dir + "primary_beam.fits",
            exposure_time_map_path=output_data_dir + "exposure_time_map.fits",
            real_visibilities_path=output_data_dir + 'real_visibilities.fits',
            imaginary_visibilities_path=output_data_dir + 'imaginary_visibilities.fits',
            visibilities_noise_map_path=output_data_dir + 'visibilities_noise_map.fits',
            u_wavelengths_path=output_data_dir + 'u_wavelengths.fits',
            v_wavelengths_path=output_data_dir + 'v_wavelengths.fits',
            renormalize_psf=False,
            renormalize_primary_beam=False,
        )

        assert (interferometer_data.image == np.ones((3, 3))).all()
        assert (interferometer_data.psf == 2.0 * np.ones((3, 3))).all()
        assert (interferometer_data.noise_map == 3.0 * np.ones((3, 3))).all()
        assert (interferometer_data.primary_beam == 5.0 * np.ones((3, 3))).all()
        assert (interferometer_data.exposure_time_map == 6.0 * np.ones((3, 3))).all()
        assert (interferometer_data.real_visibilities == np.ones(3)).all()
        assert (interferometer_data.imaginary_visibilities == 2.0 * np.ones(3)).all()
        assert (interferometer_data.visibilities_noise_map == 3.0 * np.ones(3)).all()
        assert (interferometer_data.u_wavelengths == 4.0 * np.ones(3)).all()
        assert (interferometer_data.v_wavelengths == 5.0 * np.ones(3)).all()

        assert interferometer_data.pixel_scale == 0.1
        assert interferometer_data.psf.pixel_scale == 0.1
        assert interferometer_data.noise_map.pixel_scale == 0.1
        assert interferometer_data.exposure_time_map.pixel_scale == 0.1
        


import autofit as af
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration.tests import runner

test_type = "lens_only"
test_name = "lens_x1_galaxy_hyper_background"
data_type = "lens_only_bulge_and_disk"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):

    phase1 = phase_imaging.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, light=lp.EllipticalSersic)),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    class HyperLensPlanePhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies = results.from_phase("phase_1").variable.galaxies

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            # self.hyper_image_sky = (
            #     results.last.hyper_combined.constant.hyper_image_sky
            # )

            self.hyper_noise_background = (
                results.last.hyper_combined.constant.hyper_noise_background
            )

    phase2 = HyperLensPlanePhase(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, light=lp.EllipticalSersic, hyper_galaxy=g.HyperGalaxy
            )
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration.tests import runner

test_type = "model_mapper"
test_name = "prior_limits"
data_type = "lens_only_dev_vaucouleurs"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class MMPhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens.sersic.centre_0 = 0.0
            self.galaxies.lens.sersic.centre_1 = 0.0
            self.galaxies.lens.sersic.axis_ratio = af.UniformPrior(
                lower_limit=-0.5, upper_limit=0.1
            )
            self.galaxies.lens.sersic.phi = 90.0
            self.galaxies.lens.sersic.intensity = af.UniformPrior(
                lower_limit=-0.5, upper_limit=0.1
            )
            self.galaxies.lens.sersic.effective_radius = 1.3
            self.galaxies.lens.sersic.sersic_index = 3.0

    phase1 = MMPhase(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, sersic=lp.EllipticalSersic)),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    class MMPhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens.sersic.intensity = results.from_phase(
                "phase_1"
            ).variable.lens.sersic.intensity
            self.galaxies.lens = results.from_phase("phase_1").variable.lens

    phase2 = MMPhase(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(lens=gm.GalaxyModel(redshift=0.5, sersic=lp.EllipticalSersic)),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 20
    phase2.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import numpy as np
import scipy.signal
from skimage.transform import resize, rescale

from autolens import exc
from autolens.data.array.util import array_util, grid_util, mapping_util
from autolens.data.instrument import abstract_data
from autolens.data.array import scaled_array

class InteferometerData(abstract_data.AbstractData):

    def __init__(self, image, pixel_scale, psf, primary_beam, noise_map, real_visibilities,
                 imaginary_visibilities, visibilities_noise_map, u_wavelengths, v_wavelengths, exposure_time_map=None):

        super(InteferometerData, self).__init__(image=image, pixel_scale=pixel_scale, psf=psf, noise_map=noise_map,
                                                exposure_time_map=exposure_time_map, origin=(0.0, 0.0))

        self.primary_beam = primary_beam
        self.real_visibilities = real_visibilities
        self.imaginary_visibilities = imaginary_visibilities
        self.visibilities = np.sqrt(np.square(real_visibilities) + np.square(imaginary_visibilities))
        self.visibilities_noise_map = visibilities_noise_map
        self.u_wavelengths = u_wavelengths
        self.v_wavelengths = v_wavelengths
        
    def new_interferometer_data_with_resized_arrays(
        self, new_shape, new_centre_pixels=None, new_centre_arcsec=None
    ):

        image = self.resize_scaled_array(
            scaled_array=self.image,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        noise_map = self.resize_scaled_array(
            scaled_array=self.noise_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        exposure_time_map = self.resize_scaled_array(
            scaled_array=self.exposure_time_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        return InteferometerData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            exposure_time_map=exposure_time_map,
            primary_beam=self.primary_beam,
            real_visibilities=self.real_visibilities,
            imaginary_visibilities=self.imaginary_visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            u_wavelengths=self.u_wavelengths,
            v_wavelengths=self.v_wavelengths,
        )

    def new_interferometer_data_with_resized_psf(self, new_shape):
        psf = self.resize_scaled_array(scaled_array=self.psf, new_shape=new_shape)
        return InteferometerData(
            image=self.image,
            pixel_scale=self.pixel_scale,
            psf=psf,
            noise_map=self.noise_map,
            exposure_time_map=self.exposure_time_map,
            primary_beam=self.primary_beam,
            real_visibilities=self.real_visibilities,
            imaginary_visibilities=self.imaginary_visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            u_wavelengths=self.u_wavelengths,
            v_wavelengths=self.v_wavelengths,
        )

    def new_interferometer_data_with_resized_primary_beam(self, new_shape):
        primary_beam = self.resize_scaled_array(scaled_array=self.primary_beam, new_shape=new_shape)
        return InteferometerData(
            image=self.image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=self.noise_map,
            exposure_time_map=self.exposure_time_map,
            primary_beam=primary_beam,
            real_visibilities=self.real_visibilities,
            imaginary_visibilities=self.imaginary_visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            u_wavelengths=self.u_wavelengths,
            v_wavelengths=self.v_wavelengths,
        )

    def new_interferometer_data_converted_from_electrons(self):

        image = self.array_from_counts_to_electrons_per_second(array=self.image)
        noise_map = self.array_from_counts_to_electrons_per_second(array=self.noise_map)

        return InteferometerData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            exposure_time_map=self.exposure_time_map,
            primary_beam=self.primary_beam,
            real_visibilities=self.real_visibilities,
            imaginary_visibilities=self.imaginary_visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            u_wavelengths=self.u_wavelengths,
            v_wavelengths=self.v_wavelengths,
        )

    def new_interferometer_data_converted_from_adus(self, gain):

        image = self.array_from_adus_to_electrons_per_second(
            array=self.image, gain=gain
        )
        noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.noise_map, gain=gain
        )

        return InteferometerData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            exposure_time_map=self.exposure_time_map,
            primary_beam=self.primary_beam,
            real_visibilities=self.real_visibilities,
            imaginary_visibilities=self.imaginary_visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            u_wavelengths=self.u_wavelengths,
            v_wavelengths=self.v_wavelengths,
        )


class NoiseMap(abstract_data.AbstractNoiseMap):
    
    pass


class PrimaryBeam(scaled_array.ScaledSquarePixelArray):

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale, renormalize=False, **kwargs):
        """
        Class storing a 2D Point Spread Function (PrimaryBeam), including its blurring kernel.

        Parameters
        ----------
        array : ndarray
            The 2d PrimaryBeam blurring kernel.
        renormalize : bool
            Renormalize the PrimaryBeam such that he sum of kernel values total 1.0?
        """

        # noinspection PyArgumentList
        super().__init__(array=array, pixel_scale=pixel_scale)
        if renormalize:
            self[:, :] = np.divide(self, np.sum(self))

    @classmethod
    def from_gaussian(
        cls, shape, pixel_scale, sigma, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0
    ):
        """Simulate the primary beam as an elliptical Gaussian profile."""
        from autolens.model.profiles.light_profiles import EllipticalGaussian

        gaussian = EllipticalGaussian(
            centre=centre, axis_ratio=axis_ratio, phi=phi, intensity=1.0, sigma=sigma
        )
        grid_1d = grid_util.grid_1d_from_mask_pixel_scales_sub_grid_size_and_origin(
            mask=np.full(shape, False),
            pixel_scales=(pixel_scale, pixel_scale),
            sub_grid_size=1,
        )
        gaussian_1d = gaussian.intensities_from_grid(grid=grid_1d)

        gaussian_2d = mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=gaussian_1d,
            mask=np.full(fill_value=False, shape=shape),
            sub_grid_size=1,
        )

        return PrimaryBeam(array=gaussian_2d, pixel_scale=pixel_scale, renormalize=True)

    @classmethod
    def from_fits_renormalized(cls, file_path, hdu, pixel_scale):
        """Loads a primary beam from fits and renormalizes it

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PrimaryBeam
        hdu : int
            The HDU the PrimaryBeam is stored in the .fits file.

        Returns
        -------
        primary_beam: PrimaryBeam
            A renormalized PrimaryBeam instance
        """
        primary_beam = PrimaryBeam.from_fits_with_scale(file_path, hdu, pixel_scale)
        primary_beam[:, :] = np.divide(primary_beam, np.sum(primary_beam))
        return primary_beam

    @classmethod
    def from_fits_with_scale(cls, file_path, hdu, pixel_scale):
        """
        Loads the PrimaryBeam from a .fits file.

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PrimaryBeam
        hdu : int
            The HDU the PrimaryBeam is stored in the .fits file.
        """
        return cls(
            array=array_util.numpy_array_2d_from_fits(file_path, hdu),
            pixel_scale=pixel_scale,
        )

    def new_primary_beam_with_rescaled_odd_dimensioned_array(
        self, rescale_factor, renormalize=True
    ):

        primary_beam_rescaled = rescale(
            self,
            rescale_factor,
            anti_aliasing=False,
            mode="constant",
            multichannel=False,
        )

        if primary_beam_rescaled.shape[0] % 2 == 0 and primary_beam_rescaled.shape[1] % 2 == 0:
            primary_beam_rescaled = resize(
                primary_beam_rescaled,
                output_shape=(primary_beam_rescaled.shape[0] + 1, primary_beam_rescaled.shape[1] + 1),
                anti_aliasing=False,
                mode="constant",
            )
        elif primary_beam_rescaled.shape[0] % 2 == 0 and primary_beam_rescaled.shape[1] % 2 != 0:
            primary_beam_rescaled = resize(
                primary_beam_rescaled,
                output_shape=(primary_beam_rescaled.shape[0] + 1, primary_beam_rescaled.shape[1]),
                anti_aliasing=False,
                mode="constant",
            )
        elif primary_beam_rescaled.shape[0] % 2 != 0 and primary_beam_rescaled.shape[1] % 2 == 0:
            primary_beam_rescaled = resize(
                primary_beam_rescaled,
                output_shape=(primary_beam_rescaled.shape[0], primary_beam_rescaled.shape[1] + 1),
                anti_aliasing=False,
                mode="constant",
            )

        pixel_scale_factors = (
            self.shape[0] / primary_beam_rescaled.shape[0],
            self.shape[1] / primary_beam_rescaled.shape[1],
        )
        pixel_scale = (
            self.pixel_scale * pixel_scale_factors[0],
            self.pixel_scale * pixel_scale_factors[1],
        )
        return PrimaryBeam(
            array=primary_beam_rescaled, pixel_scale=np.max(pixel_scale), renormalize=renormalize
        )

    def new_primary_beam_with_renormalized_array(self):
        """Renormalize the PrimaryBeam such that its data_vector values sum to unity."""
        return PrimaryBeam(array=self, pixel_scale=self.pixel_scale, renormalize=True)

    def convolve(self, array_2d):
        """
        Convolve an array with this PrimaryBeam

        Parameters
        ----------
        image : ndarray
            An array representing the image the PrimaryBeam is convolved with.

        Returns
        -------
        convolved_image : ndarray
            An array representing the image after convolution.

        Raises
        ------
        KernelException if either PrimaryBeam primary_beam dimension is odd
        """
        if self.shape[0] % 2 == 0 or self.shape[1] % 2 == 0:
            raise exc.KernelException("PrimaryBeam Kernel must be odd")

        return scipy.signal.convolve2d(array_2d, self, mode="same")


def load_interferometer_data_from_fits(
    image_path,
    pixel_scale,
    real_visibilities_path=None,
    imaginary_visibilities_path=None,
    visibilities_noise_map_path=None,
    u_wavelengths_path=None,
    v_wavelengths_path=None,
    image_hdu=0,
    resized_interferometer_shape=None,
    resized_interferometer_origin_pixels=None,
    resized_interferometer_origin_arcsec=None,
    psf_path=None,
    psf_hdu=0,
    resized_psf_shape=None,
    renormalize_psf=True,
    primary_beam_path=None,
    primary_beam_hdu=0,
    resized_primary_beam_shape=None,
    renormalize_primary_beam=True,
    noise_map_path=None,
    noise_map_hdu=0,
    convert_noise_map_from_weight_map=False,
    convert_noise_map_from_inverse_noise_map=False,
    exposure_time_map_path=None,
    exposure_time_map_hdu=0,
    exposure_time_map_from_single_value=None,
    real_visibilities_hdu=0,
    imaginary_visibilities_hdu=0,
    visibilities_noise_map_hdu=0,
    u_wavelengths_hdu=0,
    v_wavelengths_hdu=0,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
):
    """Factory for loading the interferometer instrument from .fits files, as well as computing properties like the noise-map,
    exposure-time map, etc. from the interferometer-instrument.

    This factory also includes a number of routines for converting the interferometer-instrument from units not supported by PyAutoLens \
    (e.g. adus, electrons) to electrons per second.

    Parameters
    ----------
    lens_name
    image_path : str
        The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
    pixel_scale : float
        The size of each pixel in arc seconds.
    image_hdu : int
        The hdu the image is contained in the .fits file specified by *image_path*.        
    image_hdu : int
        The hdu the image is contained in the .fits file that *image_path* points too.
    resized_interferometer_shape : (int, int) | None
        If input, the interferometer arrays that are image sized, e.g. the image, noise-maps) are resized to these dimensions.
    resized_interferometer_origin_pixels : (int, int) | None
        If the interferometer arrays are resized, this defines a new origin (in pixels) around which recentering occurs.
    resized_interferometer_origin_arcsec : (float, float) | None
        If the interferometer arrays are resized, this defines a new origin (in arc-seconds) around which recentering occurs.
    primary_beam_path : str
        The path to the primary_beam .fits file containing the primary_beam (e.g. '/path/to/primary_beam.fits')        
    primary_beam_hdu : int
        The hdu the primary_beam is contained in the .fits file specified by *primary_beam_path*.
    resized_primary_beam_shape : (int, int) | None
        If input, the primary_beam is resized to these dimensions.
    renormalize_psf : bool
        If True, the PrimaryBeam is renoralized such that all elements sum to 1.0.
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')        
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path to the background_noise_map .fits file containing the background noise-map \ 
        (e.g. '/path/to/background_noise_map.fits')        
    background_noise_map_hdu : int
        The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    poisson_noise_map_path : str
        The path to the poisson_noise_map .fits file containing the Poisson noise-map \
         (e.g. '/path/to/poisson_noise_map.fits')        
    poisson_noise_map_hdu : int
        The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
    poisson_noise_map_from_image : bool
        If True, the Poisson noise-map is estimated using the image.
    convert_poisson_noise_map_from_weight_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_poisson_noise_map_from_inverse_noise_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    exposure_time_map_path : str
        The path to the exposure_time_map .fits file containing the exposure time map \ 
        (e.g. '/path/to/exposure_time_map.fits')        
    exposure_time_map_hdu : int
        The hdu the exposure_time_map is contained in the .fits file specified by *exposure_time_map_path*.
    exposure_time_map_from_single_value : float
        The exposure time of the interferometer imaging, which is used to compute the exposure-time map as a single value \
        (see *ExposureTimeMap.from_single_value*).
    exposure_time_map_from_inverse_noise_map : bool
        If True, the exposure-time map is computed from the background noise_map map \
        (see *ExposureTimeMap.from_background_noise_map*)
    background_sky_map_path : str
        The path to the background_sky_map .fits file containing the background sky map \
        (e.g. '/path/to/background_sky_map.fits').
    background_sky_map_hdu : int
        The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """

    image = abstract_data.load_image(
        image_path=image_path, image_hdu=image_hdu, pixel_scale=pixel_scale
    )

    psf = abstract_data.load_psf(
        psf_path=psf_path,
        psf_hdu=psf_hdu,
        pixel_scale=pixel_scale,
        renormalize=renormalize_psf,
    )

    noise_map = load_noise_map(
        noise_map_path=noise_map_path,
        noise_map_hdu=noise_map_hdu,
        pixel_scale=pixel_scale,
        convert_noise_map_from_weight_map=convert_noise_map_from_weight_map,
        convert_noise_map_from_inverse_noise_map=convert_noise_map_from_inverse_noise_map,
    )

    exposure_time_map = abstract_data.load_exposure_time_map(
        exposure_time_map_path=exposure_time_map_path,
        exposure_time_map_hdu=exposure_time_map_hdu,
        pixel_scale=pixel_scale,
        shape=image.shape,
        exposure_time=exposure_time_map_from_single_value,
        exposure_time_map_from_inverse_noise_map=False,
        inverse_noise_map=None,
    )

    real_visibilities = load_visibilities(visibilities_path=real_visibilities_path, visibilities_hdu=real_visibilities_hdu)
    imaginary_visibilities = load_visibilities(visibilities_path=imaginary_visibilities_path, visibilities_hdu=imaginary_visibilities_hdu)
    visibilities_noise_map = load_visibilities_noise_map(visibilities_noise_map_path=visibilities_noise_map_path, visibilities_noise_map_hdu=visibilities_noise_map_hdu)
    u_wavelengths = load_visibilities(visibilities_path=u_wavelengths_path, visibilities_hdu=u_wavelengths_hdu)
    v_wavelengths = load_visibilities(visibilities_path=v_wavelengths_path, visibilities_hdu=v_wavelengths_hdu)

    primary_beam = load_primary_beam(
        primary_beam_path=primary_beam_path,
        primary_beam_hdu=primary_beam_hdu,
        pixel_scale=pixel_scale,
        renormalize=renormalize_primary_beam,
    )

    interferometer_data = InteferometerData(
        image=image,
        pixel_scale=pixel_scale,
        psf=psf,
        primary_beam=primary_beam,
        noise_map=noise_map, real_visibilities=real_visibilities, imaginary_visibilities=imaginary_visibilities,
        visibilities_noise_map=visibilities_noise_map,
        u_wavelengths=u_wavelengths, v_wavelengths=v_wavelengths, exposure_time_map=exposure_time_map,
    )

    if resized_interferometer_shape is not None:
        interferometer_data = interferometer_data.new_interferometer_data_with_resized_arrays(
            new_shape=resized_interferometer_shape,
            new_centre_pixels=resized_interferometer_origin_pixels,
            new_centre_arcsec=resized_interferometer_origin_arcsec,
        )

    if resized_psf_shape is not None:
        interferometer_data = interferometer_data.new_interferometer_data_with_resized_psf(new_shape=resized_psf_shape)

    if resized_primary_beam_shape is not None:
        interferometer_data = interferometer_data.new_interferometer_data_with_resized_primary_beam(new_shape=resized_primary_beam_shape)

    if convert_from_electrons:
        interferometer_data = interferometer_data.new_interferometer_data_converted_from_electrons()
    elif convert_from_adus:
        interferometer_data = interferometer_data.new_interferometer_data_converted_from_adus(gain=gain)

    return interferometer_data

def load_noise_map(
    noise_map_path,
    noise_map_hdu,
    pixel_scale,
    convert_noise_map_from_weight_map,
    convert_noise_map_from_inverse_noise_map,
):
    """Factory for loading the noise-map from a .fits file.

    This factory also includes a number of routines for converting the noise-map from from other units (e.g. \
    a weight map) or computing the noise-map from other unblurred_image_1d (e.g. the interferometer image and background noise-map).

    Parameters
    ----------
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    image : ndarray
        The image-image, which the noise-map can be calculated using.
    background_noise_map : ndarray
        The background noise-map, which the noise-map can be calculated using.
    exposure_time_map : ndarray
        The exposure-time map, which the noise-map can be calculated using.
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path and filename of the .fits image containing the background noise-map.
    background_noise_map_hdu : int
        The hdu the background noise-map is contained in the .fits file that *background_noise_map_path* points too.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """
    noise_map_options = sum(
        [
            convert_noise_map_from_weight_map,
            convert_noise_map_from_inverse_noise_map,
        ]
    )

    if noise_map_options > 1:
        raise exc.DataException(
            "You have specified more than one method to load the noise_map map, e.g.:"
            "convert_noise_map_from_weight_map | "
            "convert_noise_map_from_inverse_noise_map |"
            "noise_map_from_image_and_background_noise_map"
        )

    if noise_map_options == 0 and noise_map_path is not None:
        return NoiseMap.from_fits_with_pixel_scale(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scale=pixel_scale
        )
    elif convert_noise_map_from_weight_map and noise_map_path is not None:
        weight_map = scaled_array.Array.from_fits(file_path=noise_map_path, hdu=noise_map_hdu)
        return NoiseMap.from_weight_map(weight_map=weight_map, pixel_scale=pixel_scale)
    elif convert_noise_map_from_inverse_noise_map and noise_map_path is not None:
        inverse_noise_map = scaled_array.Array.from_fits(file_path=noise_map_path, hdu=noise_map_hdu)
        return NoiseMap.from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map, pixel_scale=pixel_scale
        )
    else:
        raise exc.DataException(
            "A noise_map map was not loaded, specify a noise_map_path or option to compute a noise_map map."
        )


def load_visibilities(visibilities_path, visibilities_hdu):

    if visibilities_path is not None:
        return array_util.numpy_array_1d_from_fits(file_path=visibilities_path, hdu=visibilities_hdu)


def load_visibilities_noise_map(visibilities_noise_map_path, visibilities_noise_map_hdu):
    if visibilities_noise_map_path is not None:
        return array_util.numpy_array_1d_from_fits(file_path=visibilities_noise_map_path, hdu=visibilities_noise_map_hdu)


def load_wavelengths(wavelengths_path, wavelengths_hdu):
    if wavelengths_path is not None:
        return array_util.numpy_array_1d_from_fits(file_path=wavelengths_path, hdu=wavelengths_hdu)

def load_primary_beam(primary_beam_path, primary_beam_hdu, pixel_scale, renormalize=False):
    """Factory for loading the primary_beam from a .fits file.

    Parameters
    ----------
    primary_beam_path : str
        The path to the primary_beam .fits file containing the primary_beam (e.g. '/path/to/primary_beam.fits')
    primary_beam_hdu : int
        The hdu the primary_beam is contained in the .fits file specified by *primary_beam_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    renormalize : bool
        If True, the PrimaryBeam is renoralized such that all elements sum to 1.0.
    """
    if renormalize:
        return PrimaryBeam.from_fits_renormalized(
            file_path=primary_beam_path, hdu=primary_beam_hdu, pixel_scale=pixel_scale
        )
    if not renormalize:
        return PrimaryBeam.from_fits_with_scale(
            file_path=primary_beam_path, hdu=primary_beam_hdu, pixel_scale=pixel_scale
        )
    
    
def output_interferometer_data_to_fits(
    interferometer_data,
    image_path,
    psf_path,
    noise_map_path=None,
    primary_beam_path=None,
    exposure_time_map_path=None,
        real_visibilities_path=None,
        imaginary_visibilities_path=None,
        visibilities_noise_map_path=None,
        u_wavelengths_path=None,
        v_wavelengths_path=None,
    overwrite=False,
):
    array_util.numpy_array_2d_to_fits(
        array_2d=interferometer_data.image, file_path=image_path, overwrite=overwrite
    )
    array_util.numpy_array_2d_to_fits(
        array_2d=interferometer_data.psf, file_path=psf_path, overwrite=overwrite
    )

    if primary_beam_path is not None:
        array_util.numpy_array_2d_to_fits(
            array_2d=interferometer_data.primary_beam, file_path=primary_beam_path, overwrite=overwrite
        )


    if interferometer_data.noise_map is not None and noise_map_path is not None:
        array_util.numpy_array_2d_to_fits(
            array_2d=interferometer_data.noise_map, file_path=noise_map_path, overwrite=overwrite
        )

    if interferometer_data.exposure_time_map is not None and exposure_time_map_path is not None:
        array_util.numpy_array_2d_to_fits(
            array_2d=interferometer_data.exposure_time_map,
            file_path=exposure_time_map_path,
            overwrite=overwrite,
        )
        
    if interferometer_data.real_visibilities is not None and real_visibilities_path is not None:
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.real_visibilities, file_path=real_visibilities_path, overwrite=overwrite
        )
        
    if interferometer_data.imaginary_visibilities is not None and imaginary_visibilities_path is not None:
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.imaginary_visibilities, file_path=imaginary_visibilities_path, overwrite=overwrite
        )
        
    if interferometer_data.visibilities_noise_map is not None and visibilities_noise_map_path is not None:
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.visibilities_noise_map, file_path=visibilities_noise_map_path, overwrite=overwrite
        )
        
    if interferometer_data.u_wavelengths is not None and u_wavelengths_path is not None:
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.u_wavelengths, file_path=u_wavelengths_path, overwrite=overwrite
        )
        
    if interferometer_data.v_wavelengths is not None and v_wavelengths_path is not None:
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.v_wavelengths, file_path=v_wavelengths_path, overwrite=overwrite
        )


from autolens.pipeline import pipeline_tagging

from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg


class TestPipelineNameTag:
    def test__pipeline_name_and_tag__mixture_of_values(self):

        pipeline_name = pipeline_tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl", fix_lens_light=True
        )

        assert pipeline_name == "pl_fix_lens_light"

        pipeline_tag = pipeline_tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl",
            include_shear=True,
            fix_lens_light=True,
            pixelization=pix.Rectangular,
            regularization=reg.Constant,
        )

        assert pipeline_tag == "pl_with_shear_fix_lens_light_pix_rect_reg_const"

        pipeline_name = pipeline_tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl2", fix_lens_light=True, align_bulge_disk_phi=True
        )

        assert pipeline_name == "pl2_fix_lens_light_bd_align_phi"

        pipeline_name = pipeline_tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl3",
            disk_as_sersic=True,
            align_light_dark_centre=True,
            align_bulge_dark_centre=True,
        )

        assert pipeline_name == "pl3_disk_sersic_light_dark_align_centre_bulge_dark_align_centre"

    def test__pipeline_tag__mixture_of_values(self):

        pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings(
            fix_lens_light=True
        )

        assert pipeline_tag == "_fix_lens_light"

        pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings(
            fix_lens_light=True,
            pixelization=pix.Rectangular,
            regularization=reg.Constant,
        )

        assert pipeline_tag == "_fix_lens_light_pix_rect_reg_const"

        pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings(
            fix_lens_light=True, align_bulge_disk_phi=True
        )

        assert pipeline_tag == "_fix_lens_light_bd_align_phi"

        pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings(
            align_light_dark_centre=True, align_bulge_dark_centre=True, disk_as_sersic=True
        )

        assert pipeline_tag == "_disk_sersic_light_dark_align_centre_bulge_dark_align_centre"


class TestPipelineTaggers:
    def test__fix_lens_light_tagger(self):
        tag = pipeline_tagging.fix_lens_light_tag_from_fix_lens_light(
            fix_lens_light=False
        )
        assert tag == ""
        tag = pipeline_tagging.fix_lens_light_tag_from_fix_lens_light(
            fix_lens_light=True
        )
        assert tag == "_fix_lens_light"

    def test__pixelization_tagger(self):

        tag = pipeline_tagging.pixelization_tag_from_pixelization(pixelization=None)
        assert tag == ""
        tag = pipeline_tagging.pixelization_tag_from_pixelization(
            pixelization=pix.Rectangular
        )
        assert tag == "_pix_rect"
        tag = pipeline_tagging.pixelization_tag_from_pixelization(
            pixelization=pix.VoronoiBrightnessImage
        )
        assert tag == "_pix_voro_image"

    def test__regularization_tagger(self):

        tag = pipeline_tagging.regularization_tag_from_regularization(
            regularization=None
        )
        assert tag == ""
        tag = pipeline_tagging.regularization_tag_from_regularization(
            regularization=reg.Constant
        )
        assert tag == "_reg_const"
        tag = pipeline_tagging.regularization_tag_from_regularization(
            regularization=reg.AdaptiveBrightness
        )
        assert tag == "_reg_adapt_bright"

    def test__align_bulge_disk_taggers(self):
        tag = pipeline_tagging.align_bulge_disk_centre_tag_from_align_bulge_disk_centre(
            align_bulge_disk_centre=False
        )
        assert tag == ""
        tag = pipeline_tagging.align_bulge_disk_centre_tag_from_align_bulge_disk_centre(
            align_bulge_disk_centre=True
        )
        assert tag == "_bd_align_centre"

        tag = pipeline_tagging.align_bulge_disk_axis_ratio_tag_from_align_bulge_disk_axis_ratio(
            align_bulge_disk_axis_ratio=False
        )
        assert tag == ""
        tag = pipeline_tagging.align_bulge_disk_axis_ratio_tag_from_align_bulge_disk_axis_ratio(
            align_bulge_disk_axis_ratio=True
        )
        assert tag == "_bd_align_axis_ratio"

        tag = pipeline_tagging.align_bulge_disk_phi_tag_from_align_bulge_disk_phi(
            align_bulge_disk_phi=False
        )
        assert tag == ""
        tag = pipeline_tagging.align_bulge_disk_phi_tag_from_align_bulge_disk_phi(
            align_bulge_disk_phi=True
        )
        assert tag == "_bd_align_phi"

    def test__bulge_disk_tag(self):
        tag = pipeline_tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=False,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=False,
        )
        assert tag == ""

        tag = pipeline_tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=False,
        )
        assert tag == "_bd_align_centre"

        tag = pipeline_tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=True,
        )
        assert tag == "_bd_align_centre_bd_align_phi"

        tag = pipeline_tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            align_bulge_disk_phi=True,
        )
        assert tag == "_bd_align_centre_bd_align_axis_ratio_bd_align_phi"

    def test__disk_as_sersic_tagger(self):
        tag = pipeline_tagging.disk_as_sersic_tag_from_disk_as_sersic(
            disk_as_sersic=False
        )
        assert tag == ""
        tag = pipeline_tagging.disk_as_sersic_tag_from_disk_as_sersic(
            disk_as_sersic=True
        )
        assert tag == "_disk_sersic"

    def test__align_light_dark_tagger(self):
        tag = pipeline_tagging.align_light_dark_centre_tag_from_align_light_dark_centre(
            align_light_dark_centre=False
        )
        assert tag == ""
        tag = pipeline_tagging.align_light_dark_centre_tag_from_align_light_dark_centre(
            align_light_dark_centre=True
        )
        assert tag == "_light_dark_align_centre"

    def test__align_bulge_dark_tagger(self):

        tag = pipeline_tagging.align_bulge_dark_centre_tag_from_align_bulge_dark_centre(
            align_bulge_dark_centre=False
        )
        assert tag == ""
        tag = pipeline_tagging.align_bulge_dark_centre_tag_from_align_bulge_dark_centre(
            align_bulge_dark_centre=True
        )
        assert tag == "_bulge_dark_align_centre"
